#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifdef UNIX
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

// on Solaris only
#include <sys/filio.h>
#endif // UNIX

#ifdef WIN32
#include <windows.h>
#include <winsock.h>
#endif

int laddress = 0;
int faddress = 0;
int lport = 5400;
int fport = 491;
char lhost[50];
char fhost[50];
int verbose = 0;

void	createServer();
void	runAcceptLoop(int fd);
void	sysSockets_Init();
void	sysSockets_DeInit();
void	printUsage(int fexit)
{
	printf("usage:\n");
	printf("[-lhost <localhost>] [-lport <localport>] [-fhost <forwarding host>] [-fport <forwarding port>] [-v]\n");
	if ( fexit ) exit(1);
}

int main( int argc, char *argv[] )
{
	strcpy( lhost, "0.0.0.0" );
	strcpy( fhost, "127.0.0.1" );
	int i;
	for ( i=1; i< argc; i++ )
	{
		if ( strcmp( argv[i], "-lhost") == 0 )
		{
			strcpy( lhost, argv[i+1] );
			i++;
		}
		if ( strcmp( argv[i], "-fhost") == 0 )
		{
			strcpy( fhost, argv[i+1] );
			i++;
		}
		if ( strcmp( argv[i], "-fport") == 0 )
		{
			fport = atol( argv[i+1] );
			i++;
		}
		if ( strcmp( argv[i], "-lport") == 0 )
		{
			lport = atol( argv[i+1] );
			i++;
		}
		if ( strcmp( argv[i], "-v") == 0 )
		{
			verbose = atol( argv[i+1] );
			i++;
		}
		if ( strcmp( argv[i], "-h") == 0 )
		{
			printUsage(1);
		}
		if ( strcmp( argv[i], "-help") == 0 )
		{
			printUsage(1);
		}
		if ( strcmp( argv[i], "-?") == 0 )
		{
			printUsage(1);
		}
	}

#ifdef WIN32
sysSockets_Init();
#endif
	printf( "Forward from %s:%d to %s:%d\n", 
		lhost, lport, fhost, fport );

	if ( !isdigit(lhost[0]) || !isdigit(fhost[0]) )
	{
		printf("numerical IP addresses only please!\n");
		exit(1);
	}

	laddress = ntohl( inet_addr( lhost ) );
	faddress = ntohl( inet_addr( fhost ) );

	printf( "Numerical addresses 0x%08x 0x%08x\n", lhost, fhost);

#ifdef UNIX
	signal( SIGCHLD, SIG_IGN );
#endif
	createServer();

#ifdef WIN32
sysSockets_DeInit();
#endif
	return 0;
}

void	createServer()
{
	int fd;

	struct sockaddr_in him;
	fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (fd == -1)
	{
		printf("createServer - create socket failed\n");
	}

	/* Set the reuse address on so that this port can be reused if we
	   should die without cleaning up properly. */

	int status, on;
	on = 1;
	status = setsockopt (fd, SOL_SOCKET, SO_REUSEADDR,
		(const char *)&on, sizeof (on));

	if (status == -1)
	{
		printf("createServer - SO_REUSEADDR failed\n");
	}

	/* bind */
	memset(&him, 0, sizeof him);
	him.sin_port = htons((short)lport);
	him.sin_addr.s_addr = htonl( laddress );
	him.sin_family = AF_INET;

	if (::bind (fd, (struct sockaddr *)&him, sizeof him) == -1)
	{
		printf("createServer - problem 2\n");
	}

	if (::listen(fd, 5) == -1)
	{
		printf("createServer - problem 3\n");
	}

	runAcceptLoop( fd );
}

void handleConnectionUNIX( int fd1 );
void handleConnectionWin32( int fd1 );

void	runAcceptLoop(int fd)
{
#ifdef WIN32
	int inet_addr_size;
#else
	socklen_t inet_addr_size;
#endif
	struct sockaddr_in inet_addr;
	memset( &inet_addr, 0, sizeof(inet_addr) );
	inet_addr_size = sizeof( inet_addr );
	int s;

	while ( true )
	{
		s = accept( fd, (struct sockaddr *)&inet_addr, &inet_addr_size );
		if ( s == -1 )
		{
			printf("Error accepting\n");
		}
		else
		{
			printf("Accept OK\n");
#ifdef UNIX
			int pid = fork();
			if ( pid < 0 )
			{
				printf("Error forking\n");
			}
			else if ( pid == 0 )
			{
				close( fd );
				handleConnectionUNIX( s );
				exit(0);
			}
			else
			{
				printf( "forked process pid=%d\n", pid );
				close( s );
			}
#endif
#ifdef WIN32
			handleConnectionWin32( s );
#endif
		}
	}
}

#ifdef UNIX
void handleConnectionUNIX( int fd1 )
{
	int pid = getpid();
	printf("Handle connection in pid=%d\n", pid);
	struct sockaddr_in him;
	int fd2;
	fd2 = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (fd2 == -1)
	{
		printf("problem 1\n");
	}

	int status;
	int on = 1;
	status = setsockopt (fd2, SOL_SOCKET, SO_REUSEADDR,
		(const char *)&on, sizeof (on));
	if ( status == -1  )
	{
		printf("problem 1\n");
	}

	/* connect */
	memset(&him, 0, sizeof him);
	him.sin_port = htons((unsigned short)fport);
	him.sin_addr.s_addr = htonl( faddress );
	him.sin_family = AF_INET;
	if (connect(fd2, (struct sockaddr *)&him, sizeof him) < 0)
	{
		printf("failed to connect closing\n");
		close( fd2 );
		close( fd1 );
		exit( -2 );
	}

	// here we have fd1 and fd2 connections - lets shuttle data

	printf( "connection set up L=%d F=%d\n", fd1, fd2 );
	int maxfds = fd1;
	if ( fd2>fd1 ) maxfds = fd2;
	int lastread = 0;
	int data_in[2];
	int fda[2];
	int fddir[2];
	char *data = NULL;
	int L = 0x10000;
	data = new char[L];
	bool ioerror = false;
	fda[0] = fd1;
	fda[1] = fd2;
	while( !ioerror )
	{
		fd_set fds;
		fd_set fdse;
		FD_ZERO(&fds);
		FD_ZERO(&fdse);
		FD_SET( fd1, &fdse );
		FD_SET( fd2, &fdse );
		FD_SET( fd1, &fds );
		FD_SET( fd2, &fds );
		data_in[0] = 0;
		data_in[1] = 0;
		int ret = select(maxfds+1, &fds, NULL, &fdse, NULL );
		if ( ret > 0 )
		{
			if ( FD_ISSET( fd1, &fdse ) )
			{
				printf( "L has error\n");
				ioerror = true;
			}
			if ( FD_ISSET( fd2, &fdse ) )
			{
				printf( "F has error\n");
				ioerror = true;
			}

			if ( ioerror )
				continue;

			if ( FD_ISSET( fd1, &fds ) )
			{
				if ( verbose > 0 )
				printf( "L has data\n");
				data_in[0] = 1;
			}
			if ( FD_ISSET( fd2, &fds ) )
			{
				if ( verbose > 0 )
				printf( "F has data\n");
				data_in[1] = 1;
			}

			// decide where to read from and where to write to
			fddir[0] = -1;  // read
			fddir[1] = -1;  // write
			if ( data_in[0] == 1 && data_in[1] == 1 )
			{
				if ( lastread == 1 )
				{
					fddir[0] = fda[0];
					fddir[1] = fda[1];
					lastread = 0;
				}
				else if ( lastread == 0 )
				{
					fddir[0] = fda[1];
					fddir[1] = fda[0];
					lastread = 1;
				}
			}
			else if ( data_in[0] == 1  )
			{
				fddir[0] = fda[0];
				fddir[1] = fda[1];
				lastread = 0;
			}
			else if ( data_in[1] == 1  )
			{
				fddir[0] = fda[1];
				fddir[1] = fda[0];
				lastread = 1;
			}
			else
			{
				printf("Impossible error???\n");
			}

			// first check available bytes
			int r;
			long bytes;
			r = ioctl( fddir[0], FIONREAD, &bytes );
			if ( r < 0 )
			{
				printf( "Available error on readable fd=%d\n", fddir[0] );
			}
			else
			{
				if ( bytes == 0 )
					printf( "Available 0 on readable fd=%d\n", fddir[0] );
			}

			// read and write
			int n = read( fddir[0], data, L );
			if ( n < 0 )
			{
				printf( "read error on fd=%d\n", fddir[0] );
				ioerror = true;
			}
			else if ( n == 0 )
			{
				printf( "read zero on fd=%d\n", fddir[0] );
				ioerror = true;
			}
			else
			{
				int nw = 0;
				if ( verbose > 1 )
				printf( "read %d bytes from fd=%d\n", n, fddir[0] );
				while( nw < n )
				{
					int r = write( fddir[1], &data[nw], n-nw );
					if ( r<0 )
					{
						printf( "write error to fd %d\n", fddir[1] );
						ioerror = true;
						break;
					}
					if ( verbose > 1 )
						printf( "wrote %d bytes to fd=%d\n", r, fddir[1] );
					nw += r;
				}
			}
		}
		else
		{
			printf("select error\n");
			ioerror = true;
		}
	}

	delete [] data;
	printf("Handle connection in pid=%d closing connections\n", pid);
	close( fd1 );
	close( fd2 );
	printf("Handle connection in pid=%d return\n", pid);

}
#endif // UNIX


#ifdef WIN32
typedef struct _socket_pipe {
	SOCKET fd0;
	SOCKET fd1;
} socket_pipe;

DWORD runRead( void *d )
{
	socket_pipe *sp = (socket_pipe *)d;
	SOCKET fddir[2];
	fddir[0] = sp->fd0;
	fddir[1] = sp->fd1;
	delete sp;

	int pid = GetCurrentThreadId();
	printf("IO direction thread r=%d w=%d\n", fddir[0], fddir[1]);

	char * data;
	int L = 0x10000;
	data = new char[L];
	bool ioerror = false;
	while ( !ioerror )
	{
		int n = recv( fddir[0], data, L, 0 );
		if ( n < 0 )
		{
			printf( "read error on fd=%d\n", fddir[0] );
			ioerror = true;
		}
		else if ( n == 0 )
		{
			printf( "read zero on fd=%d\n", fddir[0] );
			ioerror = true;
		}
		else
		{
			int nw = 0;
			if ( verbose > 1 )
			printf( "read %d bytes from fd=%d\n", n, fddir[0] );
			while( nw < n )
			{
				int r = send( fddir[1], &data[nw], n-nw, 0 );
				if ( r<0 )
				{
					printf( "write error to fd %d\n", fddir[1] );
					ioerror = true;
					break;
				}
				if ( verbose > 1 )
					printf( "wrote %d bytes to fd=%d\n", r, fddir[1] );
				nw += r;
			}
		}
	}

	delete [] data;

	printf("IO direction thread r=%d w=%d closing both sockets\n", fddir[0], fddir[1]);
	closesocket( fddir[0] );
	closesocket( fddir[1] );
	printf("IO direction thread r=%d w=%d return\n", fddir[0], fddir[1]);

	return 0;
}

void handleConnectionWin32( int s )
{
	SOCKET fd1 = (SOCKET)s;
	int pid = GetCurrentProcessId();
	printf("Handle connection in pid=%d\n", pid);
	struct sockaddr_in him;
	SOCKET fd2;
	fd2 = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (fd2 == -1)
	{
		printf("problem 1\n");
	}

	int status;
	int on = 1;
	status = setsockopt (fd2, SOL_SOCKET, SO_REUSEADDR,
		(const char *)&on, sizeof (on));
	if ( status == -1  )
	{
		printf("problem 1\n");
	}

	/* connect */
	memset(&him, 0, sizeof him);
	him.sin_port = htons((unsigned short)fport);
	him.sin_addr.s_addr = htonl( faddress );
	him.sin_family = AF_INET;
	if (connect(fd2, (struct sockaddr *)&him, sizeof him) < 0)
	{
		printf("failed to connect closing\n");
		closesocket( fd2 );
		closesocket( fd1 );
		return;
	}

	// here we have fd1 and fd2 connections - lets shuttle data

	printf( "connection set up L=%d F=%d\n", fd1, fd2 );

	// start 2 threads - each in one direction
	socket_pipe *sp1 = new socket_pipe;
	socket_pipe *sp2 = new socket_pipe;
	HANDLE h;
	DWORD tid;
	sp1->fd0 = fd1;
	sp1->fd1 = fd2;
	sp2->fd0 = fd2;
	sp2->fd1 = fd1;

	h = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)runRead, sp1, 0, &tid );
	CloseHandle( h );
	h = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)runRead, sp2, 0, &tid );
	CloseHandle( h );
}

void
sysSockets_Init()
{
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
	 
	wVersionRequested = MAKEWORD( 1, 1 );
	 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		/* Tell the user that we could not find a usable */
		/* WinSock DLL.                                  */
		return;
	}
	 
	 
	if ( LOBYTE( wsaData.wVersion ) != 1 ||
			HIBYTE( wsaData.wVersion ) != 1 ) {
		/* Tell the user that we could not find a usable */
		/* WinSock DLL.                                  */
		WSACleanup( );
		return;
	}
}

void
sysSockets_DeInit()
{
	WSACleanup( );
}
#endif
